﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("GQAPNMoveEvaluator", "Evaluates an n-move.")]
  [StorableClass]
  public class GQAPNMoveEvaluator : SingleSuccessorOperator, IGQAPNMoveEvaluator {

    public ILookupParameter<IntegerVector> AssignmentParameter {
      get { return (ILookupParameter<IntegerVector>)Parameters["Assignment"]; }
    }
    public ILookupParameter<NMove> MoveParameter {
      get { return (ILookupParameter<NMove>)Parameters["Move"]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<DoubleValue> FlowDistanceQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["FlowDistanceQuality"]; }
    }
    public ILookupParameter<DoubleValue> InstallationQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["InstallationQuality"]; }
    }
    public ILookupParameter<DoubleValue> OverbookedCapacityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["OverbookedCapacity"]; }
    }
    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    public ILookupParameter<DoubleValue> MoveFlowDistanceQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveFlowDistanceQuality"]; }
    }
    public ILookupParameter<DoubleValue> MoveInstallationQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveInstallationQuality"]; }
    }
    public ILookupParameter<DoubleValue> MoveOverbookedCapacityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveOverbookedCapacity"]; }
    }
    public IValueLookupParameter<DoubleValue> TransportationCostsParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["TransportationCosts"]; }
    }
    public IValueLookupParameter<DoubleValue> OverbookedCapacityPenaltyParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["OverbookedCapacityPenalty"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ILookupParameter<DoubleMatrix> InstallationCostsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["InstallationCosts"]; }
    }
    public ILookupParameter<DoubleArray> DemandsParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Demands"]; }
    }
    public ILookupParameter<DoubleArray> CapacitiesParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Capacities"]; }
    }

    [StorableConstructor]
    protected GQAPNMoveEvaluator(bool deserializing) : base(deserializing) { }
    protected GQAPNMoveEvaluator(GQAPNMoveEvaluator original, Cloner cloner) : base(original, cloner) { }
    public GQAPNMoveEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<IntegerVector>("Assignment", "The equipment-location assignment vector."));
      Parameters.Add(new LookupParameter<NMove>("Move", "The move to perform."));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The solution quality."));
      Parameters.Add(new LookupParameter<DoubleValue>("FlowDistanceQuality", "The quality regarding the flow-distance criteria."));
      Parameters.Add(new LookupParameter<DoubleValue>("InstallationQuality", "The quality regarding the installation costs."));
      Parameters.Add(new LookupParameter<DoubleValue>("OverbookedCapacity", "The sum of the overbooked capacities relative to the capacity of each location."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The quality of the move if it would be applied."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveFlowDistanceQuality", "The quality of the move regarding the flow-distance criteria."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveInstallationQuality", "The quality of the move regarding the installation costs."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveOverbookedCapacity", "The sum of the overbooked capacities of the move relative to the capacity of each location."));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("OverbookedCapacityPenalty", "The multiplier for the constraint violation when added to the quality."));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("TransportationCosts", "The transportation cost represents the flow-unit per distance-unit cost factor. This value can also be set to 1 if these costs are factored into the weights or distance matrix already."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", "The weights matrix describes the flows between the equipments."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", "The distances matrix describes the distances between the locations at which the equipment can be installed."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("InstallationCosts", "The installation costs matrix describes the installation costs of installing equipment i at location j"));
      Parameters.Add(new LookupParameter<DoubleArray>("Demands", "The demands vector describes the space requirements of the equipments."));
      Parameters.Add(new LookupParameter<DoubleArray>("Capacities", "The capacities vector describes the available space at the locations."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GQAPNMoveEvaluator(this, cloner);
    }

    public static double Evaluate(NMove move, IntegerVector assignment, DoubleMatrix weights, DoubleMatrix distances, DoubleMatrix installationCosts,
      DoubleArray demands, DoubleArray capacities, DoubleValue transportationCosts, DoubleValue overbookedCapacityPenalty) {
      double moveFlowDistanceQuality, moveInstallationQuality, moveOverbookedCapacity;
      Evaluate(move, assignment, weights, distances, installationCosts, demands, capacities,
        out moveFlowDistanceQuality, out moveInstallationQuality, out moveOverbookedCapacity);
      return GQAPEvaluator.GetCombinedQuality(moveFlowDistanceQuality, moveInstallationQuality, moveOverbookedCapacity,
        transportationCosts.Value, overbookedCapacityPenalty.Value);
    }

    public static void Evaluate(NMove move, IntegerVector assignment, DoubleMatrix weights, DoubleMatrix distances, DoubleMatrix installationCosts,
      DoubleArray demands, DoubleArray capacities, out double moveFlowDistanceQuality, out double moveInstallationQuality, out double moveOverbookedCapacity) {
      moveFlowDistanceQuality = moveInstallationQuality = moveOverbookedCapacity = 0.0;
      int moves = move.N;
      var slack = (DoubleArray)capacities.Clone();
      var oldSlack = (DoubleArray)slack.Clone();
      Dictionary<int, int> moving = new Dictionary<int, int>();
      for (int i = 0; i < moves; i++) moving.Add(move.Equipments[i], move.Locations[i]);

      for (int i = 0; i < moves; i++) {
        int equip = move.Equipments[i];
        int newLoc = move.Locations[i];
        moveInstallationQuality -= installationCosts[equip, assignment[equip]];
        moveInstallationQuality += installationCosts[equip, newLoc];
        for (int j = 0; j < assignment.Length; j++) {
          if (!moving.ContainsKey(j)) {
            moveFlowDistanceQuality += weights[equip, j] * distances[newLoc, assignment[j]];
            moveFlowDistanceQuality -= weights[equip, j] * distances[assignment[equip], assignment[j]];
            moveFlowDistanceQuality += weights[j, equip] * distances[assignment[j], newLoc];
            moveFlowDistanceQuality -= weights[j, equip] * distances[assignment[j], assignment[equip]];
            if (i == 0) { // only once for each untouched equipment deduct the demand from the capacity
              slack[assignment[j]] -= demands[j];
              oldSlack[assignment[j]] -= demands[j];
            }
          } else {
            moveFlowDistanceQuality += weights[equip, j] * distances[newLoc, moving[j]];
            moveFlowDistanceQuality -= weights[equip, j] * distances[assignment[equip], assignment[j]];
          }
        }
        slack[newLoc] -= demands[equip];
        oldSlack[assignment[equip]] -= demands[equip];
      }

      moveOverbookedCapacity = slack.Select((v, i) => new { V = v, Index = i }).Where(x => x.V < 0.0).Select(x => -x.V / capacities[x.Index]).Sum()
                               - oldSlack.Select((v, i) => new { V = v, Index = i }).Where(x => x.V < 0.0).Select(x => -x.V / capacities[x.Index]).Sum();
    }

    public override IOperation Apply() {
      double moveFlowDistanceQuality, moveInstallationQuality, moveOverbookedCapacity;
      double quality = QualityParameter.ActualValue.Value;
      double transportationCosts = TransportationCostsParameter.ActualValue.Value;
      double overbookedCapacityPenalty = OverbookedCapacityPenaltyParameter.ActualValue.Value;

      double flowDistanceQuality = FlowDistanceQualityParameter.ActualValue.Value;
      double installationQuality = InstallationQualityParameter.ActualValue.Value;
      double overbookedCapacity = OverbookedCapacityParameter.ActualValue.Value;

      Evaluate(MoveParameter.ActualValue,
               AssignmentParameter.ActualValue,
               WeightsParameter.ActualValue, DistancesParameter.ActualValue,
               InstallationCostsParameter.ActualValue,
               DemandsParameter.ActualValue, CapacitiesParameter.ActualValue,
               out moveFlowDistanceQuality, out moveInstallationQuality, out moveOverbookedCapacity);

      MoveFlowDistanceQualityParameter.ActualValue = new DoubleValue(flowDistanceQuality + moveFlowDistanceQuality);
      MoveInstallationQualityParameter.ActualValue = new DoubleValue(installationQuality + moveInstallationQuality);
      MoveOverbookedCapacityParameter.ActualValue = new DoubleValue(overbookedCapacity + moveOverbookedCapacity);

      MoveQualityParameter.ActualValue = new DoubleValue(GQAPEvaluator.GetCombinedQuality(
        flowDistanceQuality + moveFlowDistanceQuality,
        installationQuality + moveInstallationQuality,
        overbookedCapacity + moveOverbookedCapacity,
        transportationCosts,
        overbookedCapacityPenalty));
      return base.Apply();
    }
  }
}
