﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.GeneralizedQuadraticAssignment.Common;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("GreedyRandomizedSolutionCreator", "Creates a solution according to the procedure described in Mateus, G., Resende, M., and Silva, R. 2011. GRASP with path-relinking for the generalized quadratic assignment problem. Journal of Heuristics 17, Springer Netherlands, pp. 527-565.")]
  [StorableClass]
  public class GreedyRandomizedSolutionCreator : GQAPStochasticSolutionCreator {

    public IValueLookupParameter<IntValue> MaximumTriesParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumTries"]; }
    }

    [StorableConstructor]
    protected GreedyRandomizedSolutionCreator(bool deserializing) : base(deserializing) { }
    protected GreedyRandomizedSolutionCreator(GreedyRandomizedSolutionCreator original, Cloner cloner)
      : base(original, cloner) { }
    public GreedyRandomizedSolutionCreator()
      : base() {
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumTries", "The maximum number of tries to create a feasible solution.", new IntValue(10000)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GreedyRandomizedSolutionCreator(this, cloner);
    }

    protected override IntegerVector CreateRandomSolution(IRandom random, DoubleArray demands, DoubleArray capacities) {
      int tries = 0, maxTries = MaximumTriesParameter.ActualValue.Value;
      var assignment = new Dictionary<int, int>();
      var slack = new Dictionary<int, double>();

      while (tries < maxTries) {
        assignment.Clear();
        capacities.Select((v, i) => slack[i] = v).Enumerate();
        HashSet<int> CF = new HashSet<int>(), // set of chosen facilities / equipments
          T = new HashSet<int>(), // set of facilities / equpiments that can be assigned to the set of chosen locations (CL)
          CL = new HashSet<int>(), // set of chosen locations
          F = new HashSet<int>(Enumerable.Range(0, demands.Length)), // set of (initially) all facilities / equipments
          L = new HashSet<int>(Enumerable.Range(0, capacities.Length)); // set of (initially) all locations

        double threshold = 1.0;
        do {
          if (L.Any() && random.NextDouble() < threshold) {
            int l = L.ChooseRandom(random);
            L.Remove(l);
            CL.Add(l);
            T = new HashSet<int>(WithDemandEqualOrLess(F, GetMaximumSlack(slack, CL), demands));
          }
          if (T.Any()) {
            int f = T.ChooseRandom(random);
            T.Remove(f);
            F.Remove(f);
            CF.Add(f);
            int l = WithSlackGreaterOrEqual(CL, demands[f], slack).ChooseRandom(random);
            assignment.Add(f, l);
            slack[l] -= demands[f];
            T = new HashSet<int>(WithDemandEqualOrLess(F, GetMaximumSlack(slack, CL), demands));
            threshold = 1.0 - (double)T.Count / Math.Max(F.Count, 1.0);
          }
        } while (T.Any() || L.Any());
        tries++;
        if (!F.Any()) break;
      }

      if (assignment.Count != demands.Length) throw new InvalidOperationException(String.Format("No solution could be found in {0} tries.", maxTries));

      return new IntegerVector(assignment.OrderBy(x => x.Key).Select(x => x.Value).ToArray());
    }

    private static IEnumerable<int> WithDemandEqualOrLess(IEnumerable<int> facilities, double maximum, DoubleArray demands) {
      foreach (int f in facilities) {
        if (demands[f] <= maximum) yield return f;
      }
    }

    private static double GetMaximumSlack(Dictionary<int, double> slack, HashSet<int> CL) {
      return slack.Where(x => CL.Contains(x.Key)).Select(x => x.Value).Max();
    }

    private static IEnumerable<int> WithSlackGreaterOrEqual(HashSet<int> locations, double minimum, Dictionary<int, double> slack) {
      foreach (int l in locations) {
        if (slack[l] >= minimum) yield return l;
      }
    }
  }
}
